package de.brightbyte.wikiword.analyzer;

import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.brightbyte.wikiword.ConceptType;
import de.brightbyte.wikiword.Corpus;
import de.brightbyte.wikiword.Namespace;
import de.brightbyte.wikiword.ResourceType;
import de.brightbyte.wikiword.analyzer.LanguageConfiguration;
import de.brightbyte.wikiword.analyzer.PlainTextAnalyzer;
import de.brightbyte.wikiword.analyzer.WikiConfiguration;
import de.brightbyte.wikiword.analyzer.WikiTextAnalyzer;

/**
 * Unit tests for WikiTextAnalyzer
 */
public class WikiTextAnalyzerTest extends WikiTextAnalyzerTestBase {
	
	public WikiTextAnalyzerTest() {
		super("test");
	}

	//TODO: tests for all sensors (and manglers) !
	
	protected class TestWikiTextAnalyzer extends WikiTextAnalyzer {
		//TODO: check coverage!

		public TestWikiTextAnalyzer(PlainTextAnalyzer language) {
			super(language);
		}
		
		protected WikiPage makeTestPage(String title, String text) {
			return makePage(0, title, text, false);
		}
		
		public void testUnarmor() {
			TextArmor armor = new TextArmor();
			armor.put("[1]", "one");
			armor.put("[2]", "two");
			armor.put("[3]", "three");
			
			CharSequence s = unarmor("and [1] and [2] and [3] and [4]!", armor);
			assertEquals("simple unarmor", "and one and two and three and [4]!", s.toString());
			
			armor = new TextArmor();
			armor.put("[1]", "foo");
			armor.put("[2]", "bar");
			armor.put("[3]", "[1][2]");

			s = unarmor("it's [1], [2], and [3]!", armor);
			assertEquals("nested unarmor", "it's foo, bar, and foobar!", s.toString());			
		}

		public void testStripClutter() {
			TextArmor armor = new TextArmor();
			String text = "Hello [[Image:Hello.jpg|23px]] World {{FULLPAGENAME}}";
			
			CharSequence s = stripClutter(text, 0, "TEST", armor);
			assertEquals("inline image", "Hello  World TEST", s.toString());
			
  			text = "Hello\n\n[[Image:Cutaway view east rift zone Kilauea.gif|thumb|240px|Schnittzeichnung von der [[Caldera (Krater)]] des K\u012blauea \u00fcber die \u00f6stliche [[Grabenbruch|Riftzone]] mit dem [[Pu\u02bbu \u02bb\u014c\u02bb\u014d]] und durch die ''P\u016blama Pali'']]\nWorld\n{{FULLPAGENAME}}";
			s = stripClutter(text, 1, "TEST", armor);
			assertEquals("complex image with links in description", "Hello\n\n\nWorld\nTalk:TEST", s.toString());
			
			//TODO: comment, nowiki, pre, gallery, magic ...
			//TODO: nested links in images, also odd like [[Image:Hello.jpg|bla [[ foo]]
		}

		public void testStripBoxes() {
			String text = "Hello\n"
				+"{{test|foo=bar}}\n"
				+"World";
			
			CharSequence s = stripBoxes(text);
			assertEquals("lone template", "Hello\n\nWorld", s.toString());
			
			//TODO: tables, div, center, nesting, complex, broken
		}

		public void testStripMarkup() {
			String text;
			CharSequence s;
			String t;
						
			text = "* Interwiki:-[[w:Foo]]-\n" +
			"* Langlink:-[[fr:Foo]]-\n" +
			"* Namespace:-[[User:Foo|blubb]]-\n" +
			"* Colon:-[[Say: what?]]-\n" +
			"* Category:-[[Category:Foo]]-\n" +
			"* Plain:-[[Foo|blah]]-\nx";
			s = stripMarkup(text);

			t = "* Interwiki:--\n" +
			"* Langlink:--\n" +
			"* Namespace:--\n" +
			"* Colon:-Say: what?-\n" +
			"* Category:--\n" +
			"* Plain:-blah-\nx";
			assertEquals("qualified links", t, s.toString());
			
			text = "Hello [[World]]!";
			s = stripMarkup(text);
			assertEquals("simple link", "Hello World!", s.toString());
			
			text = "Foo &amp; Bar &stuff!";
			s = stripMarkup(text);
			assertEquals("html entities", "Foo & Bar &stuff!", s.toString());
			
			text = "Check '''''this''' out'' dude!";
			s = stripMarkup(text);
			assertEquals("emphasis", "Check this out dude!", s.toString());
			
			text = "Some <b>bold</b> stuff!";
			s = stripMarkup(text);
			assertEquals("html tags", "Some bold stuff!", s.toString());
			
			text = "Entities: &Uuml;, &#64;, &#x0040; &&;!";
			s = stripMarkup(text);
			assertEquals("html entities", "Entities: \u00dc, @, @ &&;!", s.toString());
			
			//TODO: hr, indent, bullets, etc
		}

		public void testDetermineResourceType() {
			WikiPage page = testAnalyzer.makeTestPage("Foo", "#REDIRECT [[Bar]]<!-- go there! -->");
			ResourceType pt = determineResourceType(page);
			assertSame("redirect", ResourceType.REDIRECT, pt);

			page = testAnalyzer.makeTestPage("Foo (Disambiguation)", "yadda yadda");
			pt = determineResourceType(page);
			assertSame("disambig (title)", ResourceType.DISAMBIG, pt);

			page = testAnalyzer.makeTestPage("Foo", "{{disambiguation}}\nyadda");
			pt = determineResourceType(page);
			assertSame("disambig (template)", ResourceType.DISAMBIG, pt);

			page = testAnalyzer.makeTestPage("Foo", "{{delete}}\nyadda");
			pt = determineResourceType(page);
			assertSame("bad (delete)", ResourceType.BAD, pt);

			page = testAnalyzer.makeTestPage("Foo", "yadda yadda");
			pt = determineResourceType(page);
			assertSame("plain article", ResourceType.ARTICLE, pt);

			page = testAnalyzer.makeTestPage("Foo", "yadda\n#REDIRECT\nyadda");
			pt = determineResourceType(page);
			assertSame("article like redirect", ResourceType.ARTICLE, pt);
		}

		public void testDetermineConceptType() {
			WikiPage page = testAnalyzer.makeTestPage("Foo", "yadda [[Category:Died 1956]]");
			ConceptType ct = determineConceptType(page);
			assertSame("person", ConceptType.PERSON, ct);

			page = testAnalyzer.makeTestPage("Foo", "yadda [[Category:Town in Hubbub]]");
			ct = determineConceptType(page);
			assertSame("place", ConceptType.PLACE, ct);

			page = testAnalyzer.makeTestPage("Foo", "yadda {{Taxobox\n|foo=bar}}");
			ct = determineConceptType(page);
			assertSame("lifeform", ConceptType.LIFEFORM, ct);

			page = testAnalyzer.makeTestPage("1866", "bla bla");
			ct = determineConceptType(page);
			assertSame("time", ConceptType.TIME, ct);

			page = testAnalyzer.makeTestPage("Foobar", "bla bla");
			ct = determineConceptType(page);
			assertSame("time", ConceptType.OTHER, ct);
		}

		public void testDetermineTitleSuffix() {
			String title = "Talk:Foo (Bar)"; 
			CharSequence s = determineTitleSuffix(title);
			
			assertEquals("Bar", s.toString());
		}

		public void testDetermineTitleTerms() {
			Set<String> exp = new HashSet<String>();
			exp.add("Foo");
			exp.add("foo");
			
			WikiPage page = testAnalyzer.makeTestPage("Foo", "{{DISPLAYTITLE:foo}} yadda yadda");
			Set<String> terms = determineTitleTerms(page);
			assertEquals("plain title", exp, terms);
		}

		public void testExtractRedirectLink() {
			String text = "#REDIREcT [[bar]][[Category:Orf]]";
			WikiTextAnalyzer.WikiLink link = extractRedirectLink("Foo", text);

			assertEquals("Bar", link.getPage());
		}

		public void testIsInterlanguagePrefix() {
			assertEquals("nl", true, isInterlanguagePrefix("nl"));
			assertEquals("zh-yue", true, isInterlanguagePrefix("zh-yue"));
			assertEquals("talk", false, isInterlanguagePrefix("talk"));
			assertEquals("project_talk", false, isInterlanguagePrefix("project_talk"));
			assertEquals("s", false, isInterlanguagePrefix("s"));
			assertEquals("meta", false, isInterlanguagePrefix("meta"));
		}

		public void testIsInterwikiPrefix() {
			assertEquals("nl", true, isInterwikiPrefix("nl"));
			//assertEquals("talk", false, isInterwikiPrefix("talk")); //NOTE: can't be detected here, namespace index must be checked first.
			assertEquals("project_talk", false, isInterwikiPrefix("project_talk"));
			assertEquals("s", true, isInterwikiPrefix("s"));
			assertEquals("meta", true, isInterwikiPrefix("meta"));
		}

		public void testGetNamespaceId() {
			//TODO: custom namespaces!
			assertEquals("(main)", Namespace.MAIN, getNamespaceId(""));
			assertEquals("talk", Namespace.TALK, getNamespaceId("talk"));
			assertEquals("User", Namespace.USER, getNamespaceId("User"));
			assertEquals("Template_talk", Namespace.TEMPLATE_TALK, getNamespaceId("Template_talk"));
			assertEquals("project talk", Namespace.PROJECT_TALK, getNamespaceId("project talk"));
		}

		public void testNormalizeTitle() {
			String title = " a&b c";
			CharSequence s = normalizeTitle(title);
			assertEquals("A&b_c", s.toString());
			
			//TODO: more...
		}

		public void testIsBadLinkTarget() {
			assertEquals("foo xyzzy", false, isBadLinkTarget("foo xyzzy"));
			assertEquals("(empty)", true, isBadLinkTarget(""));
			assertEquals("{{foo}}", true, isBadLinkTarget("{{foo}}"));
			assertEquals("{{{foo}}}", true, isBadLinkTarget("barf_{{{foo}}}"));
			assertEquals("barf <magic>foo</magic>", true, isBadLinkTarget("barf <magic>foo</magic>"));
			assertEquals("x|y", true, isBadLinkTarget("x|y"));
			assertEquals("..", true, isBadLinkTarget(".."));
			assertEquals("http://dummy.org/link", true, isBadLinkTarget("http://dummy.org/link"));
			assertEquals("[[bla]]x", true, isBadLinkTarget("[[bla]]x"));
			assertEquals("(tab)", true, isBadLinkTarget("bla\tblubb"));
			assertEquals("(null)", true, isBadLinkTarget("bla\0blubb"));
			assertEquals("(newline)", true, isBadLinkTarget("bla\nblubb"));
			assertEquals("(space)", false, isBadLinkTarget("bla blubb"));
			assertEquals("(umlaut)", false, isBadLinkTarget("bl\u00f6h"));
		}

		public void testIsBadTerm() {
			assertEquals("(empty)", true, isBadTerm(""));
			assertEquals("(tab)", true, isBadTerm("bla\tblubb"));
			assertEquals("(null)", true, isBadTerm("bla\0blubb"));
			assertEquals("(newline)", true, isBadTerm("bla\nblubb"));
			assertEquals("(space)", false, isBadTerm("bla blubb"));
			assertEquals("(umlaut)", false, isBadTerm("bl\u00f6h"));
		}

		public void testExtractFirstParagraph() {
			String text;
			CharSequence s;
			
			text = "{{eeuwbox|eeuw=24}}\n\n" +
					"{{eeuwoverzicht|eeuw=24}}\n\n" +
					"[[Categorie:24e eeuw| ]]\n\n" +
					"[[stq:24. Jierhunnert]]";
			
			s = extractFirstParagraph(text);
			assertNull("no text, just cruft", s);
			
			text = "\n"
				+"First thing\n"
				+"\n"
				+"end.\n";
			
			s = extractFirstParagraph(text);
			assertEquals("simple paragraph", "First thing", s.toString());
			
			text = "== section ==\n"
				+"no intro\n"
				+"end.\n";
			
			s = extractFirstParagraph(text);
			assertEquals("no paragraph", "", s.toString());
			
			text = "{{head}}\n"
				//+"[[Image:something.png|thumb]]\n" //NOTE: usually stripped as clutter
				+"<div>boxy thingy</div>\n"
				+"This is [[it]],\n"
				+"you know.\n"
				+"\n"
				+"end.\n";
			
			s = extractFirstParagraph(text);
			assertEquals("hidden paragraph", "This is it,\nyou know.", s.toString());
		}

		public void testExtractFirstSentence() {
			String text = "Foo (also abc. cde.) is the 2. [[Quux]] in [[Xyzzy]]. Its not a [[Barf]].\n";
			
			CharSequence s = language.extractFirstSentence( stripMarkup( text ) );
			assertEquals("simple sentence", "Foo is the 2. Quux in Xyzzy.", s.toString());
			
			//TODO: all the nasty stuff...
		}

		public void testExtractDisambigLinks() {
			String text = "\n{{Disambiguation}}"
				+"'''Foo''' [[may]] be:\n"
				+"* A [[bla]]\n"
				+"* A [[Yoxo]] or [[foxo]] in [[baar]]\n"
				+"== Other ==\n"
				+"* [[Math]]:\n"
				+"*# A [[quux]]\n"
				+"*# A barf: see [[info:barf]]\n"
				+"* Internal [[Project:Stuff|stuff]]\n"
				+"== See Also ==\n"
				+"* see [[xxx]]\n"
				+"* and [[xxx|xyz]]\n"
				+"[[Category:TLA]]\n"
				+"end\n";
			
			List<WikiLink> exp = new ArrayList<WikiLink>();
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "Bla", null, "bla", true, LinkMagic.NONE));
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "Foxo", null, "foxo", true, LinkMagic.NONE));
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "Quux", null, "quux", true, LinkMagic.NONE));

			WikiPage page = testAnalyzer.makeTestPage("Foo", text);
			List<WikiLink> links = extractDisambigLinks(page.getTitle(), page.getCleanedText(true));
			assertEquals("simple disambig", exp, links);
		}

		public void testExtractLinks() {
			String text;
			List<WikiLink> exp;
			WikiPage page;
			List<WikiLink> links;

			text = "";
			exp = new ArrayList<WikiLink>();
			text += "Foo [[bar]]s!\n";
			text += "check [[this|that]] out, [[simple thing]]\n";
			text += "this [[pipe | pipes]], this [[ blow|blows ]]\n";
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "Bar", null, "bars", true, LinkMagic.NONE));
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "This", null, "that", false, LinkMagic.NONE));
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "Simple_thing", null, "simple thing", true, LinkMagic.NONE));
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "Pipe", null, "pipes", false, LinkMagic.NONE));
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "Blow", null, "blows", false, LinkMagic.NONE));
			page = testAnalyzer.makeTestPage("Foo", text);
			links = extractLinks(page.getTitle(), page.getCleanedText(true));
			assertEquals(exp, links);

			text = "";
			exp = new ArrayList<WikiLink>();
			text += "12[[inch|\"]] or more\n";
			text += "[[first]] and [[:last]]\n";
			text += "[[give me some space|some space| and time]]\n";
			text += "[[odd#|stuff>]]\n";
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "Inch", null, "\"", false, LinkMagic.NONE));
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "First", null, "first", true, LinkMagic.NONE));
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "Last", null, "last", true, LinkMagic.NONE));
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "Give_me_some_space", null, "some space| and time", false, LinkMagic.NONE));
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "Odd", null, "stuff>", false, LinkMagic.NONE));
			page = testAnalyzer.makeTestPage("Foo", text);
			links = extractLinks(page.getTitle(), page.getCleanedText(true));
			assertEquals(exp, links);

			text = "";
			exp = new ArrayList<WikiLink>();
			text += "[[{{bad}}]]\n";
			text += "[[bad{bar]]\n";
			text += "[[bad}bar]]\n";
			text += "[[too ''bad'']]\n";
			text += "[[bad[bar]]\n";
			text += "[[bad]bar]]\n";
			text += "[[bad<bar]]\n";
			text += "[[bad>bar]]\n";
			text += "[[broken\nlink]]\n";			
			text += "[[this|''works'' {{too}}]]\n";
			text += "[[quite'ok']]\n";
			text += "[[section# link thing...]]\n";
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "This", null, "works {{too}}", false, LinkMagic.NONE));
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "Quite'ok'", null, "quite'ok'", true, LinkMagic.NONE));
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "Section", "link_thing...", "section# link thing...", true, LinkMagic.NONE));
			page = testAnalyzer.makeTestPage("Foo", text);
			links = extractLinks(page.getTitle(), page.getCleanedText(true));
			assertEquals(exp, links);

			text = "";
			exp = new ArrayList<WikiLink>();
			text += "[[section#.C3.84.C.ASX.Y.0B.26.05.4]]\n"; //encoded special chars in section (including .05 and .0B which should get dropped)
			text += "[[%C3%84%C%ASX%Y%0b%26%05%4]]\n"; //encoded special chars in target (including %05 and %0b which should get dropped)
			text += "[[URL%23Encoding]]\n"; //url-encoded link (yes the # may also be encoded, this does not act as an escape)
			text += "[[HTML&amp;entities]]\n"; //html-entities
			text += "[[no%special&stuff]]\n"; //no special stuff
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "Section", "\u00c4.C.ASX.Y&.4", "section#.C3.84.C.ASX.Y.0B.26.05.4", true, LinkMagic.NONE)); 
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "\u00c4%C%ASX%Y&%4", null, "\u00c4%C%ASX%Y&%4", true, LinkMagic.NONE)); 
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "URL", "Encoding", "URL#Encoding", true, LinkMagic.NONE)); 
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "HTML&entities", null, "HTML&entities", true, LinkMagic.NONE)); 
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.MAIN, "No%special&stuff", null, "no%special&stuff", true, LinkMagic.NONE)); 
			page = testAnalyzer.makeTestPage("Foo", text);
			links = extractLinks(page.getTitle(), page.getCleanedText(true));
			assertEquals(exp, links);

			exp = new ArrayList<WikiLink>();
			text = "";
			text += "\nimage: [[Image:test.jpg]], [[Image:test.jpg|thumb]], [[Image:test.jpg|the [[test]] image]], [[Image:test.jpg|the {{test}} image]];"; //NOTE: stripped as clutter
			text += "namespace: [[User:foo]], [[User talk:foo|talk]], [[:User:foo]]bar;\n";			
			exp.add(new WikiLink(null, Namespace.USER, "Foo", null, "User:foo", true, LinkMagic.NONE));
			exp.add(new WikiLink(null, Namespace.USER_TALK, "Foo", null, "talk", false, LinkMagic.NONE));
			exp.add(new WikiLink(null, Namespace.USER, "Foo", null, "User:foobar", true, LinkMagic.NONE));
			page = testAnalyzer.makeTestPage("Foo", text);
			links = extractLinks(page.getTitle(), page.getCleanedText(true));
			assertEquals(exp, links);

			text = "";
			exp = new ArrayList<WikiLink>();
			text += "[[Category:Foo| ]]\n"; //category blank sortkey
			text += "[[Category:Foo]]\n"; //category
			text += "[[:Category:Foo|Bar]]\n"; //category link
			text += "[[Category:Foo|Bar]]\n"; //category sortkey
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.CATEGORY, "Foo", null, "", false, LinkMagic.CATEGORY)); 
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.CATEGORY, "Foo", null, "Foo", true, LinkMagic.CATEGORY)); 
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.CATEGORY, "Foo", null, "Bar", false, LinkMagic.NONE)); 
			exp.add(new WikiTextAnalyzer.WikiLink(null, Namespace.CATEGORY, "Foo", null, "Bar", false, LinkMagic.CATEGORY)); 
			page = testAnalyzer.makeTestPage("Foo", text);
			links = extractLinks(page.getTitle(), page.getCleanedText(true));
			assertEquals(exp, links);
			
			exp = new ArrayList<WikiLink>();
			text = "category: [[Category: z]], [[Category: z|zz]], [[:Category: z]], [[:Category: z|z]];\n";
			exp.add(new WikiLink(null, Namespace.CATEGORY, "Z", null, "Foo", true, LinkMagic.CATEGORY));
			exp.add(new WikiLink(null, Namespace.CATEGORY, "Z", null, "zz", false, LinkMagic.CATEGORY));
			exp.add(new WikiLink(null, Namespace.CATEGORY, "Z", null, "Category: z", true, LinkMagic.NONE));
			exp.add(new WikiLink(null, Namespace.CATEGORY, "Z", null, "z", false, LinkMagic.NONE));
			page = testAnalyzer.makeTestPage("Foo", text);
			links = extractLinks(page.getTitle(), page.getCleanedText(true));
			assertEquals(exp, links);

			text = "";
			exp = new ArrayList<WikiLink>();
			text += "[[x_z:zeug]]\n"; //bad (should be caught by badLinkTarget)
			text += "[[xyz:zeug|zeug]]\n"; //interwiki
			text += "[[de:Zeug]]\n"; //interlanguage
			text += "[[:de:Zeug]]\n"; //interwiki
			exp.add(new WikiTextAnalyzer.WikiLink("xyz", Namespace.MAIN, "Zeug", null, "zeug", false, LinkMagic.NONE)); 
			exp.add(new WikiTextAnalyzer.WikiLink("de", Namespace.MAIN, "Zeug", null, "de:Zeug", true, LinkMagic.LANGUAGE)); 
			exp.add(new WikiTextAnalyzer.WikiLink("de", Namespace.MAIN, "Zeug", null, "de:Zeug", true, LinkMagic.NONE)); 
			page = testAnalyzer.makeTestPage("Foo", text);
			links = extractLinks(page.getTitle(), page.getCleanedText(true));
			assertEquals(exp, links);
			
			exp = new ArrayList<WikiLink>();
			text = "language: [[nl: z]], [[zh-yue: z|z]], [[:de: z|z]];\n";
			exp.add(new WikiLink("nl", Namespace.MAIN, "Z", null, "nl: z", true, LinkMagic.LANGUAGE));
			exp.add(new WikiLink("zh-yue", Namespace.MAIN, "Z", null, "z", false, LinkMagic.LANGUAGE));
			exp.add(new WikiLink("de", Namespace.MAIN, "Z", null, "z", false, LinkMagic.NONE));
			page = testAnalyzer.makeTestPage("Foo", text);
			links = extractLinks(page.getTitle(), page.getCleanedText(true));
			assertEquals(exp, links);
			
			exp = new ArrayList<WikiLink>();
			text = "interwiki: [[ixy: z]], [[ixy: z|z]], [[:ixy: z|z]];\n";
			exp.add(new WikiLink("ixy", Namespace.MAIN, "Z", null, "ixy: z", true, LinkMagic.NONE));
			exp.add(new WikiLink("ixy", Namespace.MAIN, "Z", null, "z", false, LinkMagic.NONE));
			exp.add(new WikiLink("ixy", Namespace.MAIN, "Z", null, "z", false, LinkMagic.NONE));
			page = testAnalyzer.makeTestPage("Foo", text);
			links = extractLinks(page.getTitle(), page.getCleanedText(true));
			assertEquals(exp, links);
			
			exp = new ArrayList<WikiLink>();
			text = "prefix: [[x y: z]], [[x y: z|z ]];\n";
			exp.add(new WikiLink(null, Namespace.MAIN, "X_y:_z", null, "x y: z", true, LinkMagic.NONE));
			exp.add(new WikiLink(null, Namespace.MAIN, "X_y:_z", null, "z", false, LinkMagic.NONE));
			page = testAnalyzer.makeTestPage("Foo", text);
			links = extractLinks(page.getTitle(), page.getCleanedText(true));
			assertEquals(exp, links);
		}

		public void testExtractTemplates() {
			String text = "something {{simple thing}} or {{trivial |x y}} or {{easy | y z | foo=23}}.\n"
				+"now {{nasty stuff|\n"
				+"dummy|\n"
				+"foo {{bar}} x|\n"
				+"foo {{bar{{!}}x=y}}|\n"
				+"one=1|\n"
				+"|two=2|three=3\n"
				+"|ugly={{thingy{{!}}3}} |\n"
				+"|{{x}}fugly= barf |\n"
				+"1=yummy\n"
				+"}}{{ping}}\n"
				+"plus {{#if:cruft|then|that}}\n"
				+"plus {{DEFAULTSORTKEY:x}}\n"
				+"end.\n";
			Map<String, Map<String, String>> exp = new HashMap<String, Map<String,String>>();
			Map<String, String> params;			
			
			params = new HashMap<String, String>();
			exp.put("Simple_thing", params);
			
			params = new HashMap<String, String>();
			params.put("1", "x y");
			exp.put("Trivial", params);

			params = new HashMap<String, String>();
			params.put("1", "y z");
			params.put("foo", "23");
			exp.put("Easy", params);

			params = new HashMap<String, String>();
			params.put("1", "yummy");
			params.put("2", "foo  x");
			params.put("3", "foo");
			params.put("one", "1");
			params.put("4", "");
			params.put("two", "2");
			params.put("three", "3");
			params.put("ugly", "");
			params.put("5", "");
			params.put("fugly", "barf");
			exp.put("Nasty_stuff", params);

			params = new HashMap<String, String>();
			exp.put("Ping", params);

			params = new HashMap<String, String>();
			params.put("0", "x"); 
			exp.put("DEFAULTSORT", params);

			Map<String, Map<String, CharSequence>> templates = extractTemplates(text);
			assertEquals("templates", exp, templates);
		}

		public void testExtractSections() {
			Set<String> sections = extractSections("");
			assertEquals(theSet(), sections);			
			
			sections = extractSections("== foo ==");
			assertEquals(theSet("foo"), sections);			

			sections = extractSections("== foo ==\n\n== bar ==\n== baz ==\n== zoosh ==\n== orf ==");
			assertEquals(theSet("foo", "bar", "baz", "zoosh", "orf"), sections);			

			sections = extractSections("xxx\n== foo ==\nxxx\n== bar ==\nxxx");
			assertEquals(theSet("foo", "bar"), sections);			

			sections = extractSections("== baz\nxxx\n== foo ==\nxxx\n=== bar ===\nxxx\n=== frob==");
			assertEquals(theSet("foo", "bar", "= frob"), sections);			
		}
		
		public void testStripSections() {
			Matcher strip = WikiConfiguration.sectionPattern("see also|external links|references", Pattern.CASE_INSENSITIVE).matcher(""); 

			String text = "";
			String expected = text;
			CharSequence stripped = stripSections(text, strip);
			assertEquals(expected, stripped.toString());			
			
			text = "foo\n\nbar\n\n== xxx ==\n\nxxx";
			expected = text;
			stripped = stripSections(text, strip);
			assertEquals(expected, stripped.toString());			
			
			text = "foo\n\nbar\n\n== xxx ==\n\nxxx";
			expected = text;
			stripped = stripSections(text, strip);
			assertEquals(expected, stripped.toString());			
			
			text = "foo\n\n" +
					"== see also ==\n\n" +
					"aaa\n\n" +
					"bbb";
			expected = "foo\n";
			stripped = stripSections(text, strip);
			assertEquals(expected, stripped.toString());			
			
			text = "foo\n\n" +
					"== xxx ==\n\n" +
					"xxx\n\n" +
					"== see also ==\n\n" +
					"aaa\n\n" +
					"=== boof ===\n\n" +
					"bbb\n\n" +
					"= zzzz =\n\n" +
					"zzz\n\n" +
					"== external links ==\n\n" +
					"eeee\n\n" +
					"== yyy ==\n\n" +
					"yyy";
			expected = "foo\n\n" +
					"== xxx ==\n\n" +
					"xxx\n\n" +
					"= zzzz =\n\n" +
					"zzz\n\n" +
					"== yyy ==\n\n" +
					"yyy";
			stripped = stripSections(text, strip);
			assertEquals(expected, stripped.toString());			
		}
		
	} 
	
	protected static <T> List<T> theList(T... x) {
		return Arrays.asList(x);
	}
	
	protected static <T> Set<T> theSet(T... x) {
		return new HashSet<T>( Arrays.asList(x) );
	}
	
	protected TestWikiTextAnalyzer testAnalyzer;
	
	@Override
	public void setUp() throws URISyntaxException {		
		LanguageConfiguration lconfig = new LanguageConfiguration();

		corpus = new Corpus("TEST", "generic", "generic", "generic", "generic", "xx", "generic");
		PlainTextAnalyzer language = new PlainTextAnalyzer(corpus);
		language.configure(lconfig);
		language.initialize();

		WikiConfiguration config = new WikiConfiguration();
		config.resourceTypeSensors.add( new WikiTextAnalyzer.RegularExpressionTitleSensor(ResourceType.DISAMBIG, "_\\(Disambiguation\\)$", 0) );
		config.resourceTypeSensors.add( new WikiTextAnalyzer.HasTemplateSensor(ResourceType.DISAMBIG, "Disambiguation", null) );
		config.resourceTypeSensors.add( new WikiTextAnalyzer.HasTemplateSensor(ResourceType.BAD, "Delete", null) );

		config.conceptTypeSensors.add( new WikiTextAnalyzer.HasCategoryLikeSensor(ConceptType.PERSON, "^Died_|^Born_", 0) );
		config.conceptTypeSensors.add( new WikiTextAnalyzer.HasCategoryLikeSensor(ConceptType.NAME, "(^N|_n)ames?$", 0) );
		config.conceptTypeSensors.add( new WikiTextAnalyzer.HasCategoryLikeSensor(ConceptType.PLACE, "^Town_in_|^Country_in_|^Region_in_", 0) );
		config.conceptTypeSensors.add( new WikiTextAnalyzer.HasTemplateSensor(ConceptType.LIFEFORM, "Taxobox", null) );
		config.conceptTypeSensors.add( new WikiTextAnalyzer.RegularExpressionTitleSensor(ConceptType.TIME, "^-?\\d+$", 0) );
		config.conceptTypeSensors.add( new WikiTextAnalyzer.HasSectionSensor(ConceptType.PERSON, "Life") );
		
		config.disambigStripSectionPattern = WikiConfiguration.sectionPattern("see also|external links", Pattern.CASE_INSENSITIVE);
		
		testAnalyzer = new TestWikiTextAnalyzer(language);
		testAnalyzer.configure(config);
		testAnalyzer.initialize(namespaces, titleCase);
		
		analyzer = testAnalyzer;
	}

	public void testDetermineConceptType() {
		testAnalyzer.testDetermineConceptType();
	}

	public void testDetermineTitleTerms() {
		testAnalyzer.testDetermineTitleTerms();
	}

	public void testDetermineResourceType() {
		testAnalyzer.testDetermineResourceType();
	}

	public void testDetermineTitleSuffix() {
		testAnalyzer.testDetermineTitleSuffix();
	}

	public void testExtractDisambigLinks() {
		testAnalyzer.testExtractDisambigLinks();
	}

	public void testExtractFirstParagraph() {
		testAnalyzer.testExtractFirstParagraph();
	}

	public void testExtractFirstSentence() {
		testAnalyzer.testExtractFirstSentence();
	}

	public void testExtractLinks() {
		testAnalyzer.testExtractLinks();
	}

	public void testExtractRedirectLink() {
		testAnalyzer.testExtractRedirectLink();
	}

	public void testExtractTemplates() {
		testAnalyzer.testExtractTemplates();
	}

	public void testGetNamespaceId() {
		testAnalyzer.testGetNamespaceId();
	}

	public void testIsBadLinkTarget() {
		testAnalyzer.testIsBadLinkTarget();
	}

	public void testIsBadTerm() {
		testAnalyzer.testIsBadTerm();
	}

	public void testIsInterlanguagePrefix() {
		testAnalyzer.testIsInterlanguagePrefix();
	}

	public void testIsInterwikiPrefix() {
		testAnalyzer.testIsInterwikiPrefix();
	}

	public void testNormalizeTitle() {
		testAnalyzer.testNormalizeTitle();
	}

	public void testStripBoxes() {
		testAnalyzer.testStripBoxes();
	}

	public void testStripClutter() {
		testAnalyzer.testStripClutter();
	}

	public void testStripMarkup() {
		testAnalyzer.testStripMarkup();
	}

	public void testUnarmor() {
		testAnalyzer.testUnarmor();
	}

	public void testExtractSections() {
		testAnalyzer.testExtractSections();
	}

	public void testStripSections() {
		testAnalyzer.testStripSections();
	}

	public static void main(String[] args) {
		run(WikiTextAnalyzerTest.class, args); 
	}
}
